searchNodes=[{"ref":"QueryElf.html","title":"QueryElf","type":"behaviour","doc":"Defines an Ecto query builder. Here&#39;s a simple usage example: defmodule MyQueryBuilder do use QueryElf, schema: MySchema, searchable_fields: [:id, :name] def filter(:my_extra_filter, value, _query) do dynamic([s], s.some_field - ^value == 0) end end MyQueryBuilder.build_query(id__in: [1, 2, 3], my_extra_filter: 10) It accepts the following options: :schema - the Ecto.Schema for which the queries will be built (required) :searchable_fields - a list of fields to build default filters for (default: []) :sortable_fields - a list of fields to build default sorters for (default: []) :plugins - a list of plugins that can be used to increment the query builder&#39;s functionality. (default: []) If used, the :searchable_fields option will define the following filters (according to the field type in the schema): id or binary_id: :$FIELD - checks if the field is equal to the given value :$FIELD__neq - checks if the field is different from the given value :$FIELD__in - checks if the field is contained in the given enumerable boolean: :$FIELD - checks if the field is equal to the given value integer, float or decimal: :$FIELD - checks if the field is equal to the given value :$FIELD__neq - checks if the field is different from the given value :$FIELD__in - checks if the field is contained in the given enumerable :$FIELD__gt - checks if the field is greater than the given value :$FIELD__lt - checks if the field is lower than the given value :$FIELD__gte - checks if the field is greater than or equal to the given value :$FIELD__lte - checks if the field is lower than or equal to the given value string: :$FIELD - checks if the field is equal to the given value :$FIELD__neq - checks if the field is different from the given value :$FIELD__in - checks if the field is contained in the given enumerable :$FIELD__contains - checks if the field contains the given string :$FIELD__starts_with - checks if the field starts with the given string :$FIELD__ends_with - checks if the field ends with the given string date, time, naive_datetime, datetime, time_usec, naive_datetime_usec or datetime_usec: :$FIELD - checks if the field is equal to the given value :$FIELD__neq - checks if the field is different from the given value :$FIELD__in - checks if the field is contained in the given enumerable :$FIELD__after - checks if the field occurs after the given value :$FIELD__before - checks if the field occurs before the given value"},{"ref":"QueryElf.html#c:__query_builder__/1","title":"QueryElf.__query_builder__/1","type":"callback","doc":"Should return the query builder metadata requested."},{"ref":"QueryElf.html#c:base_query/0","title":"QueryElf.base_query/0","type":"callback","doc":"Should return the base query in this query builder. This base query will be used when defining the empty_query/0 and build_query/1 callbacks. If not defined, defaults to the supplied schema. This is useful when dealing with logical deletion or other business rules that need to be followed every time the query builder is used. Example: defmodule UsersQueryBuilder do use QueryElf, schema: User, searchable_fields: [:id, :name] def base_query do from u in User, where: is_nil(u.deleted_at) end end"},{"ref":"QueryElf.html#c:build_query/1","title":"QueryElf.build_query/1","type":"callback","doc":"Should receive a a keyword list or a map containing parameters and use it to build a query."},{"ref":"QueryElf.html#c:build_query/2","title":"QueryElf.build_query/2","type":"callback","doc":"Same thing as build_query/1, but also receives some options for things like pagination and ordering."},{"ref":"QueryElf.html#c:build_query/3","title":"QueryElf.build_query/3","type":"callback","doc":"The same thing as build_query/2, but instead of building a new query it receives and extends an existing one."},{"ref":"QueryElf.html#c:empty_query/0","title":"QueryElf.empty_query/0","type":"callback","doc":"Should return a query that when applied, returns nothing."},{"ref":"QueryElf.html#c:filter/3","title":"QueryElf.filter/3","type":"callback","doc":"Receives an atom representing a filter, the parameter for the aforementioned filter and an Ecto query. Should return an Ecto dynamic or a tuple containing an Ecto query and an Ecto dynamic. Most of the filtering should happen in the returned dynamic, and returning a query should only be used when the query needs to be extended for the dynamic to make sense. Example: # this is a simple comparison, so just returning a dynamic will suffice def filter(:my_filter, value, _query) do dynamic([s], s.some_field == ^value) end # this relies on a join, so the query must be extended accordingly def filter(:my_other_filter, value, query) do { join(query, :left, [s], assoc(s, :some_relationship), as: :related), dynamic([related: r], r.some_field == ^value) } end"},{"ref":"QueryElf.html#reusable_join/5","title":"QueryElf.reusable_join/5","type":"macro","doc":"Similar to Ecto.Query.join/{4,5}, but can be called multiple times with the same alias. Note that only the first join operation is performed, the subsequent ones that use the same alias are just ignored. Also note that because of this behaviour, its mandatory to specify an alias when using this function. This is helpful when you need to perform a join while building queries one filter at a time, because the same filter could be used multiple times or you could have multiple filters that require the same join. This scenario poses a problem with how the filter/3 callback work, as you need to return a dynamic with the filtering, which means that the join must have an alias, and by default Ecto raises an error when you add multiple joins with the same alias. To solve this, it is recommended to use this macro instead of the default Ecto.Query.join/{4,5}. As an added bonus, there will be only one join in the query that can be reused by multiple filters."},{"ref":"QueryElf.html#c:sort/4","title":"QueryElf.sort/4","type":"callback","doc":"Receives an atom representing a field to order by, the order direction, an extra argument to perform the ordering, and an Ecto query. Should return the Ecto query with the appropriate sorting options applied. Example: def sort(:name, direction, _arg, query) do sort(query, [s], [{^direction, s.name}]) end"},{"ref":"QueryElf.html#t:filter/0","title":"QueryElf.filter/0","type":"type","doc":""},{"ref":"QueryElf.html#t:metadata_type/0","title":"QueryElf.metadata_type/0","type":"type","doc":""},{"ref":"QueryElf.html#t:options/0","title":"QueryElf.options/0","type":"type","doc":""},{"ref":"QueryElf.html#t:sort_direction/0","title":"QueryElf.sort_direction/0","type":"type","doc":""},{"ref":"QueryElf.Plugin.html","title":"QueryElf.Plugin","type":"behaviour","doc":"A behaviour to be followed when creating plugins that extends the query builders functionality."},{"ref":"QueryElf.Plugin.html#c:build_query/3","title":"QueryElf.Plugin.build_query/3","type":"callback","doc":"Takes the query, the builder, and the options given to the build_query function build the query as arguments. Should return the modified query."},{"ref":"basic_usage.html","title":"Basic Usage","type":"extras","doc":"Basic Usage Suppose you have a users schema that is defined as below: defmodule Accounts.User do schema &quot;users&quot; do field :email, :string field :password_hash, :string field :name, :string field :is_active, :boolean field :signed_up_at, :naive_datetime end end"},{"ref":"basic_usage.html#basic-filtering","title":"Basic Usage - Basic filtering","type":"extras","doc":"Suppose you wish to enable filtering by email, name, is_active, and signed_up_at. Then you can define a QueryBuilder module as below: defmodule Accounts.User.QueryBuilder do use QueryElf, schema: Accounts.User, searchable_fields: [:email, :name, :is_active, :signed_up_at] end The schema specifies the ecto schema for which the query builder is being created. The searchable_fields specifies fields that can be filtered on. With the above definition, you can now use the query builder&#39;s capabailities as below: User.QueryBuilder.build_query(email: &quot;a@b.com&quot;) |&gt; Repo.one User.QueryBuilder.build_query(name__in: [&quot;Raphael Costa&quot;, &quot;Jose Valim&quot;]) |&gt; Repo.all() User.QueryBuilder.build_query(is_active: true) |&gt; Repo.all() User.QueryBuilder.build_query(signed_up_at__before: ~N[2000-01-01 23:00:07]) |&gt; Repo.all() You can also combine multiple conditions/filters together: # Conditions are `anded` by default User.QueryBuilder.build_query(name__in: [&quot;Raphael Costa&quot;, &quot;Jose Valim&quot;], is_active: true) |&gt; Repo.all() # But you can change to using `or` User.QueryBuilder.build_query(_or: [name: &quot;Mr A&quot;, email: &quot;a@b.com&quot;]) |&gt; Repo.all() # Or combine them together to create a complex query User.QueryBuilder.build_query( _or: [ name: &quot;Mr A&quot;, _and: [ email__in: [&quot;a@b.com&quot;, &quot;a@c.com&quot;], is_active: true ] ] ) |&gt; Repo.all() Information about what filters (__in, __before, etc) are defined for each field type is specified in the module docs."},{"ref":"basic_usage.html#basic-ordering-sorting","title":"Basic Usage - Basic ordering/sorting","type":"extras","doc":"Assuming you have the same user schema defined above, suppose you wish to allow sorting users by name, email, and signed_up_at. QueryElf can be used to achieve this as below: defmodule Accounts.User.QueryBuilder do use QueryElf, schema: Accounts.User, searchable_fields: ~w[email name is_active signed_up_at]a, sortable_fields: ~w[email name signed_up_at]a end Then sorting capabilities can be included in build_query as below: Accounts.User.QueryBuilder.build_query([], order: [%{field: :name, direction: :asc}, %{field: :signed_up_at, direction: :desc}] ) |&gt; Repo.all() # Shorthand expression for ordering example above Accounts.User.QueryBuilder.build_query([], order: [asc: :my_int, desc: :my_bool] ) |&gt; Repo.all()"},{"ref":"advanced_usage.html","title":"Advanced Usage","type":"extras","doc":"Advanced Usage Suppose you have a users schema that is defined as below: defmodule Accounts.User do schema &quot;users&quot; do field :email, :string field :password_hash, :string field :name, :string field :is_active, :boolean field :signed_up_at, :naive_datetime field :last_logged_in_at, :naive_datetime end end Then a basic filtering and ordering query builder can be defined as: defmodule Accounts.User.QueryBuilder do use QueryElf, schema: Accounts.User, searchable_fields: ~w[email name is_active signed_up_at]a, sortable_fields: ~w[email name signed_up_at]a end Where filtering is possible on email, name, is_active, and signed_up_at, and sorting is possible on email, name, and signed_up_at. If this is confusing for you, it would be a good idea to read through the basic usage documentation."},{"ref":"advanced_usage.html#defining-custom-filters","title":"Advanced Usage - Defining custom filters","type":"extras","doc":"Suppose you wish to implement a custom filter to identify recently active users defined as users that logged in in the last n days. You can define and use such a filter as below: defmodule Accounts.User.QueryBuilder do use QueryElf, schema: Accounts.User, searchable_fields: ~w[email name is_active signed_up_at]a, sortable_fields: ~w[email name signed_up_at]a def filter(:active_in_last_days, days, _query) do earliest_logged_in_at = NaiveDateTime.add( NaiveDateTime.utc_now(), days * 24 * 3600 * -1, :second ) dynamic([u], u.last_logged_in_at &gt; ^earliest_logged_in_at) end end Accounts.User.QueryBuilder.build_query(active_in_last_days: 30) |&gt; Repo.all() Suppose you also have a notification preferences schema and a foriegn key from users schema to the preferences schema and you want to implement custom filters to identify recently active users that are happy to receive newsletters, then that can be achieved as below: defmodule Accounts.User do schema &quot;users&quot; do field :email, :string field :password_hash, :string field :name, :string field :is_active, :boolean field :signed_up_at, :naive_datetime field :last_logged_in_at, :naive_datetime has_one :notification_preferences, NotificationPreferences end end defmodule Accounts.NotificationPreferences do schema &quot;notification_preferences&quot; do field :send_product_updates, :boolean field :send_newsletters, :boolean end end defmodule Accounts.User.QueryBuilder do use QueryElf, schema: Accounts.User, searchable_fields: ~w[email name is_active signed_up_at]a, sortable_fields: ~w[email name signed_up_at]a def filter(:active_in_last_days, days, _query) do earliest_logged_in_at = NaiveDateTime.add( NaiveDateTime.utc_now(), days * 24 * 3600 * -1, :second ) dynamic([u], u.last_logged_in_at &gt; ^earliest_logged_in_at) end def filter(:accepts_newsletters, value, query) do { # Use reusable_join instead of join as explained in the next section join(query, assoc(r, :notification_preferences), as: :preferences), dynamic([preferences: p], p.send_newsletters == ^value) } end end Accounts.User.QueryBuilder.build_query(active_in_last_days: 30, accepts_newsletters: true) |&gt; Repo.all()"},{"ref":"advanced_usage.html#using-reusable-joins","title":"Advanced Usage - Using reusable joins","type":"extras","doc":"Given the module definition of User.QueryBuilder above, if you wrote the following query: Accounts.User.QueryBuilder.build_query( _or: [ accepts_newsletters: true, _and: [active_in_last_days: 5, accepts_newsletters: false] ] ) Then ecto with throw an error as it would result in two named joins having the same name. This is because both accepts_newsletters: true and accepts_newsletters: false add a join each between users and notification_preferences with the name preferences and ecto doesn&#39;t allow two joins with the same name. What you really want is that a single join is added to the query as they both are essentially the same join. This is achieved by the reusable_join macro. import QueryElf defmodule Accounts.User.QueryBuilder do use QueryElf, schema: Accounts.User, searchable_fields: ~w[email name is_active signed_up_at]a, sortable_fields: ~w[email name signed_up_at]a def filter(:accepts_newsletters, value, query) do { reusable_join(query, assoc(r, :notification_preferences), as: :preferences), dynamic([preferences: p], p.send_newsletters == ^value) } end end Internally, the reusable_join macro checks if a join with the same name has already been added to the query and if so it skips adding the join. Thus, if you have two custom filters that should be using the same join, you can ensure that happens by giving the joins in the filters the same name. For example, suppose you want to implement custom filters to allow fetching users that are happy to receive newsletters and product updates. In this case, you would need to build two different filters so that they can be used independently and also be combined. But you want to ensure that when combined, the filters use a single join instead of multiple joins on users and notification_preferences. This is achieved by using reusable_join with the same join name in both the custom filters. defmodule Accounts.User.QueryBuilder do use QueryElf, schema: Accounts.User, searchable_fields: ~w[email name is_active signed_up_at]a, sortable_fields: ~w[email name signed_up_at]a def filter(:accepts_newsletters, value, query) do { reusable_join(query, assoc(r, :notification_preferences), as: :preferences), dynamic([preferences: p], p.send_newsletters == ^value) } end def filter(:accepts_product_updates, value, query) do { reusable_join(query, assoc(r, :notification_preferences), as: :preferences), dynamic([preferences: p], p.send_product_updates == ^value) } end end Accounts.User.QueryBuilder.build_query(accepts_newsletters: true) |&gt; Repo.all() Accounts.User.QueryBuilder.build_query(accepts_product_updates: true) |&gt; Repo.all() Accounts.User.QueryBuilder.build_query(accepts_newsletters: true, accepts_product_updates: true) |&gt; Repo.all()"}]